const seek = value => () => value

export const Seek = Object.freeze({
    Left: seek(-1),
    Right: seek(1),
    Stay: seek(0),
    First: reader => -reader.position,
    Last: reader => (reader.length - 1) - reader.position,
    JumpFromStack: reader => {
        const state = reader.pop()
        reader.push()
        return state.position - reader.position
    },
    SeekFromStack: consuming => reader => {
        const position = (consuming ? reader.peek() : reader.pop()).position
        return position - reader.position
    }
})

export const Predicate = Object.freeze({
    AnyOf: (...symbols) => b => symbols.includes(b.symbol),
    Every: (...fns) => x => fns.every(fn => fn(x)),
    Equals: a => b => a === b.symbol,
    False: () => false,
    InBounds: direction => reader => reader.inBounds(reader.position + direction()),
    Not: fn => x => !fn(x),
    Number: x => !isNaN(x.symbol),
    OutOfBounds: direction => reader => !Predicate.InBounds(direction)(reader),
    Some: (...fns) => x => fns.some(fn => fn(x)),
    True: () => true,
})

const tee = fn => x => {
    fn(x)
    return x.symbol
}

export const Replace = Object.freeze({
    Seq: (...fns) => x => fns.reduce((_acc, fn) => fn(x), x.symbol),
    With: x => () => x,
    Nothing: x => x.symbol,
    Push: x => {
        x.push()
        return x.symbol
    },
    Pop: x => x.pop().symbol,
    Swap: x => {
        const a = x.pop()
        const b = x.pop()
        x.push(a)
        x.push(b)
        return b.symbol
    },
    ClearStack: tee(x => x.clear()),
    FromStack: x => x.peek().symbol,
    Tee: tee
})


class Head {
    #position = 0

    get position() { return this.#position }

    left() { this.move(Left) }
    right() { this.move(Right) }
    move(value) { this.#position += value }


    read(tape) {
        return tape.at(this.#position)
    }

    write(tape, value) {
        tape.write(this.#position, value)
    }
}

export class Tape {
    #value

    constructor(input) {
        this.#value = Array.isArray(input) ? input : input.split('')
    }

    get value() { return this.#value }
    get length() { return this.#value.length }

    at(n) {
        return this.value[n]
    }

    write(n, value) {
        this.#value[n] = value
    }
}

export class Rule {
    constructor(from, to, predicate, replaceSymbol, action) {
        this.fromState = from
        this.toState = to
        this.predicate = predicate
        this.replaceSymbol = replaceSymbol
        this.action = action
    }

    static is(from, to, currentSymbol, nextSymbol, action) {
        return new Rule(
            from,
            to,
            Predicate.Equals(currentSymbol),
            nextSymbol,
            action
        )
    }
}

export class Rules {
    #firstState
    #rules = new Map()

    constructor(rules) {
        this.#firstState = rules[0]
        this.#initialize(rules)
    }

    get firstState() { return this.#firstState }

    #initialize(rules) {
        rules.forEach(rule => {
            const transition = this.#rules.get(rule.fromState)

            if (transition) {
                transition.push(rule)
            } else {
                this.#rules.set(rule.fromState, [rule])
            }
        })
    }

    get(state, reader) {
        return this.#rules.get(state)?.find(rule => rule.predicate(reader)) ?? false
    }
}

class ReaderState {
    constructor(position, symbol) {
        this.position = position
        this.symbol = symbol
    }

    static from(reader) {
        return new ReaderState(reader.position, reader.symbol)
    }
}

class Iterator extends Array {
    #index = 0

    done() {
        return this.length === 0 || this.#index === this.length - 1
    }

    peek() {
        return this[this.#index]
    }

    next() {
        return this[this.#index++]
    }

    take(n) {
        const result = []
        while (!this.done() && n--) {
            result.push(this.next())
        }
        return result
    }

    skip(n) {
        this.take(n)
        return this
    }

    takeUntil(fn) {
        const result = []
        let c = 0
        while (!this.done() && this.peek() && !fn(this.peek())) {
            result.push(this.next())
            c++
        }
        return result
    }

    skipUntil(fn) {
        this.takeUntil(fn)
        return this
    }
}

export class Reader {
    #head
    #tape
    #stack = []

    constructor(head, tape) {
        this.#head = head
        this.#tape = tape
    }

    get head() { return this.#head }
    get tape() { return this.#tape }
    get stack() { return Iterator.from(this.#stack) }

    get position() {
        return this.#head.position
    }

    get length() {
        return this.#tape.length
    }

    get symbol() {
        return this.#head.read(this.#tape)
    }

    clear() {
        this.#stack.length = 0
    }

    inBounds(index) {
        return index >= 0 && index <= this.length - 1
    }

    push(x = ReaderState.from(this)) {
        this.#stack.push(x)
    }

    pop() {
        return this.#stack.pop()
    }

    peek() {
        return this.#stack[this.#stack.length - 1]
    }
}

const colors = {
    brightWhite: [97, 107],
    brightCyan: [96, 106],
    brightMagenta: [95, 105],
    brightBlue: [94, 104],
    brightYellow: [93, 103],
    brightGreen: [92, 102],
    brightRed: [91, 101],
    brightBlack: [90, 100],
    white: [37, 47],
    cyan: [36, 46],
    magenta: [35, 45],
    blue: [34, 44],
    yellow: [33, 43],
    green: [32, 42],
    red: [31, 41],
    black: [30, 40],
}

const color = (code, t) => {
    if (Array.isArray(code)) {
        return code.reduce((acc, c) => color(c, acc), t)
    }

    return `\x1b[${code}m${t}` + (t.endsWith(reset) ? '' : reset)
}
const reset = '\x1b[0m'
//const color = t => `\x1b[44m\x1b[93m${t}\x1b[0m`
const highlight = (i, arr, col = colors.cyan, spaces = 0) => {
    const a = arr.slice()
    const len = a[i].length
    a[i] = color(col, a[i])
    const diff = a[i].length - len
    //console.log(a[i].length, len, diff, spaces)
    a[i] = a[i].padEnd(spaces + diff)
    return a
}

const range = n => [...Array(n).keys()]

export class TuringMachine {
    #head = new Head()
    #rules
    #state
    #reader

    constructor(tape, rules, initialState = rules.firstState.fromState, debug = false) {
        this.tape = tape
        this.debug = debug

        this.#rules = rules
        this.#state = initialState

        this.#reader = new Reader(this.#head, this.tape)
    }

    get state() { return this.#state }
    get value() { return this.tape.value }

    display(rule, next) {
        console.log(this.#state, '->', rule.toState, '(', this.#reader.symbol, '->', next, ')')
        const length = this.tape.length
        const spaces = length.toString().length + 1
        console.log(highlight(this.#reader.position, this.value, [1, 2, 4, colors.black[1], colors.yellow[0]], spaces).join(' '.repeat(spaces)))
        const numbers = range(this.tape.length).map(String)
        const highlighted = highlight(this.#reader.position, numbers, [1, 7], spaces + 3).map(x => x.padEnd(spaces + 1, ' ')).join('')
        console.log(highlighted)
    }

    step(rule) {
        const next = rule.replaceSymbol(this.#reader)

        if (this.debug) {
            this.display(rule, next)
        }

        if (next) {
            this.#head.write(this.tape, next)
        }
        this.#state = rule.toState
        this.#head.move(rule.action(this.#reader))
    }

    next() {
        return this.#rules.get(this.#state, this.#reader)
    }

    exec() {
        let next
        while (next = this.next()) {
            this.step(next)
        }
    }
}


